# -*- coding: utf-8 -*-
"""
Created on Sun Feb  9 18:15:06 2025

@author:Prof. Heitor S.Lopes com suporte do Gemini 2.0 Flash
"""
#Algoritmo genético para o problema da mochila - versão 1
#
#Arquivos de entrada:
# produtos.txt --> Dados do problema
#   ProdutoA,10,5,10,20
#   ProdutoB,5,15,5,30
#   ProdutoC,20,2,2,5
#   ProdutoD,8,8,8,15
#   ProdutoX,20,1,1,5
# parametros.txt --> Parâmetros do Algoritmo Genético
#   pop,200
#   mut,0.1
#   cros,0.8
#   ger,100
#

import random
import numpy as np
import matplotlib.pyplot as plt

# Função para ler os dados dos produtos de um arquivo
def ler_dados_produtos(nome_arquivo):
    produtos = []
    with open(nome_arquivo, 'r') as arquivo:
        for linha in arquivo:
            nome, quantidade, peso, volume, valor = linha.strip().split(',')
            produtos.append({
                'nome': nome,
                'quantidade': int(quantidade),
                'peso': float(peso),
                'volume': float(volume),
                'valor': float(valor)
                })
    return produtos


# Função para ler os parâmetros do algoritmo genético de um arquivo
def ler_parametros_geneticos(nome_arquivo):
    parametros = {}
    with open(nome_arquivo, 'r') as arquivo:
        for linha in arquivo:
            chave, valor = linha.strip().split(',')
            parametros[chave] = float(valor) if chave in ('mut', 'cros') else int(valor)
    return parametros


# Função para criar uma população inicial de soluções
def criar_populacao(tamanho_populacao, produtos):
    populacao = []
    for _ in range(tamanho_populacao):
        solucao = {}
        for produto in produtos:
            solucao[produto['nome']] = random.randint(0, produto['quantidade'])
            populacao.append(solucao)
    return populacao

# Função para calcular a aptidão de uma solução
def calcular_aptidao(solucao, produtos, peso_maximo, volume_maximo):
    peso_total = 0
    volume_total = 0
    valor_total = 0
    for produto in produtos:
        peso_total += solucao[produto['nome']] * produto['peso']
        volume_total += solucao[produto['nome']] * produto['volume']
        valor_total += solucao[produto['nome']] * produto['valor']
 
    if peso_total > peso_maximo or volume_total > volume_maximo:
        return 0  # Solução inválida
    else:
        return valor_total  # Solução válida

# Função para realizar a seleção por torneio
def selecao_torneio(populacao, produtos, peso_maximo, volume_maximo, tamanho_torneio):
    novos_pais = []
    for _ in range(len(populacao)):
        participantes = random.sample(populacao, tamanho_torneio)
        vencedor = max(participantes, key=lambda solucao: calcular_aptidao(solucao, produtos, peso_maximo, volume_maximo))
        novos_pais.append(vencedor)
    return novos_pais

# Função para realizar o cruzamento de dois pontos
def cruzamento_dois_pontos(pai1, pai2, produtos):
    pontos_corte = sorted(random.sample(range(len(produtos)), 2))
    filho1 = pai1.copy()
    filho2 = pai2.copy()
    nomes_produtos = list(pai1.keys())  # Obtém a lista de nomes dos produtos
 
    for i in range(pontos_corte[0], pontos_corte[1]):
        nome_produto = nomes_produtos[i]  # Obtém o nome do produto na posição i
        filho1[nome_produto] = pai2[nome_produto]
        filho2[nome_produto] = pai1[nome_produto]
    return filho1, filho2

# Função para realizar a mutação simples
def mutacao_simples(solucao, produtos, taxa_mutacao):
    for produto in produtos:
        if random.random() < taxa_mutacao:
            solucao[produto['nome']] = random.randint(0, produto['quantidade'])
    return solucao
 

#---------------------------------------------------

# Função principal do algoritmo genético
def algoritmo_genetico(nome_arquivo_produtos, nome_arquivo_parametros, peso_maximo, volume_maximo):
    global produtos  # Indica que a variável global produtos será usada
    produtos = ler_dados_produtos(nome_arquivo_produtos)
    parametros = ler_parametros_geneticos(nome_arquivo_parametros)
    populacao = criar_populacao(parametros['pop'], produtos)
    melhores_aptidoes = []  # Lista para armazenar as melhores aptidões de cada geração
    medias_aptidoes = [] # Nova lista para armazenar as médias
    for geracao in range(parametros['ger']):
        # Seleção
        pais = selecao_torneio(populacao, produtos, peso_maximo, volume_maximo, 3)  # Torneio de tamanho 3
        # Cruzamento
        nova_populacao = []
        for i in range(0, len(pais), 2):
            pai1 = pais[i]
            pai2 = pais[i + 1]
            filho1, filho2 = cruzamento_dois_pontos(pai1, pai2, produtos)
            nova_populacao.extend([filho1, filho2])
        # Mutação
        for i in range(len(nova_populacao)):
            nova_populacao[i] = mutacao_simples(nova_populacao[i], produtos, parametros['mut'])
        populacao = nova_populacao
        # Encontra a melhor solução e sua aptidão na geração atual
        melhor_solucao = max(populacao, key=lambda solucao: calcular_aptidao(solucao, produtos, peso_maximo, volume_maximo))
        melhor_solucao_geracao = max(populacao, key=lambda solucao: calcular_aptidao(solucao, produtos, peso_maximo, volume_maximo))
        melhor_aptidao_geracao = calcular_aptidao(melhor_solucao_geracao, produtos, peso_maximo, volume_maximo)
        melhores_aptidoes.append(melhor_aptidao_geracao)  # Armazena a melhor aptidão da geração atual
 
        # Encontrar a melhor solução geral (após todas as gerações)
        melhor_solucao = max(populacao, key=lambda solucao: calcular_aptidao(solucao, produtos, peso_maximo, volume_maximo))
        melhor_aptidao = calcular_aptidao(melhor_solucao, produtos, peso_maximo, volume_maximo)
        
        # Calcula a média das aptidões da população
        aptidoes_geracao = [calcular_aptidao(individuo, produtos, peso_maximo, volume_maximo) for individuo in populacao]
        media_aptidao_geracao = np.mean(aptidoes_geracao)
        medias_aptidoes.append(media_aptidao_geracao) # Armazena a média na lista
 
    # Plotar a evolução da melhor aptidão ao longo das gerações
    #plt.plot(range(1, parametros['ger'] + 1), medias_aptidoes) # Plota as médias
    plt.plot(range(1, parametros['ger'] + 1), melhores_aptidoes, label='Melhor Fitness')  # Plota a melhor aptidão
    plt.plot(range(1, parametros['ger'] + 1), medias_aptidoes, label='Média dos Fitness')  # Plota a média da aptidão
 
    plt.xlabel('Geração')
    plt.ylabel('Fitness')
    plt.title('Evolução do Fitness ao Longo das Gerações')
    plt.legend()  # Adiciona a legenda para identificar as curvas
    plt.show()
        
    return melhor_solucao, melhor_aptidao


#---------------------------------------------------

# Programa principal
nome_arquivo_produtos = 'produtos.txt'
nome_arquivo_parametros = 'parametros.txt'
peso_maximo = 200
volume_maximo = 200

print("\n\n Algoritmo Genético Simples para o Problema da Mochila")

melhor_solucao, melhor_aptidao = algoritmo_genetico(nome_arquivo_produtos, nome_arquivo_parametros, peso_maximo, volume_maximo)
  

print("Melhor solução encontrada:")
for produto, quantidade in melhor_solucao.items():
    print(f"{produto}: {quantidade}")
print(f"Melhor fitness de toda população: {melhor_aptidao}")

# Calcular peso e volume totais da melhor solução
peso_total = 0
volume_total = 0
for produto_selecionado in produtos:
    peso_total += melhor_solucao[produto_selecionado['nome']] * produto_selecionado['peso']
    volume_total += melhor_solucao[produto_selecionado['nome']] * produto_selecionado['volume']
 

print(f"Peso total da carga: {peso_total}")
print(f"Volume total da carga: {volume_total}")

  